/* -*- Mode: C; indent-tabs-mode: nil; tab-width: 8; c-basic-offset: 2 -*- */

/*
 * Gnome-Mahjongg pile creation algorithm
 *
 * (C) 2003 the Free Software Foundation
 *
 * Author: Callum McKenzie, <callum@physics.otago.ac.nz>
 *
 * This code is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#include <gnome.h>
/* #include <stdlib.h> */

#include <games-files.h>

#include "mahjongg.h"
#include "maps.h"

/* FIXME: Delete most of these once they are migrated to external files.
 * (But keep easy_map as a default. */

/* Sorted such that the bottom leftest are first, and layers decrease
 * Bottom left = high y, low x ! */

tilepos easy_map [MAX_TILES] = {
 {13, 7,  4}, {12, 8,  3}, {14, 8,  3}, {12, 6,  3},
 {14, 6,  3}, {10, 10,  2}, {12, 10,  2}, {14, 10,  2},
 {16, 10,  2}, {10, 8,  2}, {12, 8,  2}, {14, 8,  2},
 {16, 8,  2}, {10, 6,  2}, {12, 6,  2}, {14, 6,  2},
 {16, 6,  2}, {10, 4,  2}, {12, 4,  2}, {14, 4,  2},
 {16, 4,  2}, {8, 12,  1}, {10, 12,  1}, {12, 12,  1},
 {14, 12,  1}, {16, 12,  1}, {18, 12,  1}, {8, 10,  1},
 {10, 10,  1}, {12, 10,  1}, {14, 10,  1}, {16, 10,  1}, 
 {18, 10,  1}, {8, 8,  1}, {10, 8,  1}, {12, 8,  1},
 {14, 8,  1}, {16, 8,  1}, {18, 8,  1}, {8, 6,  1},
 {10, 6,  1}, {12, 6,  1}, {14, 6,  1}, {16, 6,  1},
 {18, 6,  1}, {8, 4,  1}, {10, 4,  1}, {12, 4,  1},
 {14, 4,  1}, {16, 4,  1}, {18, 4,  1}, {8, 2,  1},
 {10, 2,  1}, {12, 2,  1}, {14, 2,  1}, {16, 2,  1},
 {18, 2,  1}, {2, 14,  0}, {4, 14,  0}, {6, 14,  0},
 {8, 14,  0}, {10, 14,  0}, {12, 14,  0}, {14, 14,  0}, 
 {16, 14,  0}, {18, 14,  0}, {20, 14,  0}, {22, 14,  0},
 {24, 14,  0}, {6, 12,  0}, {8, 12,  0}, {10, 12,  0},
 {12, 12,  0}, {14, 12,  0}, {16, 12,  0}, {18, 12,  0},
 {20, 12,  0}, {4, 10,  0}, {6, 10,  0}, {8, 10,  0},
 {10, 10,  0}, {12, 10,  0}, {14, 10,  0}, {16, 10,  0},
 {18, 10,  0}, {20, 10,  0}, {22, 10,  0}, {0, 7,  0},
 {2, 8,  0}, {4, 8,  0}, {6, 8,  0}, {8, 8,  0},
 {10, 8,  0}, {12, 8,  0}, {14, 8,  0}, {16, 8,  0},
 {18, 8,  0}, {20, 8,  0}, {22, 8,  0}, {24, 8,  0},
 {2, 6,  0}, {4, 6,  0}, {6, 6,  0}, {8, 6,  0},
 {10, 6,  0}, {12, 6,  0}, {14, 6,  0}, {16, 6,  0},
 {18, 6,  0}, {20, 6,  0}, {22, 6,  0}, {24, 6,  0},
 {4, 4,  0}, {6, 4,  0}, {8, 4,  0}, {10, 4,  0},
 {12, 4,  0}, {14, 4,  0}, {16, 4,  0}, {18, 4,  0},
 {20, 4,  0}, {22, 4,  0}, {6, 2,  0}, {8, 2,  0},
 {10, 2,  0}, {12, 2,  0}, {14, 2,  0}, {16, 2,  0},
 {18, 2,  0}, {20, 2,  0}, {2, 0,  0}, {4, 0,  0},
 {6, 0,  0}, {8, 0,  0}, {10, 0,  0}, {12, 0,  0},
 {14, 0,  0}, {16, 0,  0}, {18, 0,  0}, {20, 0,  0},
 {22, 0,  0}, {24, 0,  0}, {26, 7, 0}, {28, 7, 0} 
};

tilepos hard_map [MAX_TILES] = {
	{ 10, 6,  6},
	{ 9, 6,  5},
	{ 11, 6,  5},
	{ 8, 6,  4},
	{ 10, 6,  4},
	{ 12, 6,  4},
	{ 5, 6,  3},
	{ 7, 7,  3},
	{ 7, 5,  3},
	{ 9, 7,  3},
	{ 9, 5,  3},
	{ 11, 7,  3},
	{ 11, 5,  3},
	{ 13, 7,  3},
	{ 13, 5,  3},
	{ 15, 6,  3},
	{ 5, 8,  2},
	{ 7, 8,  2},
	{ 9, 8,  2},
	{ 11, 8,  2},
	{ 13, 8,  2},
	{ 15, 8,  2},
	{ 4, 6,  2},
	{ 6, 6,  2},
	{ 8, 6,  2},
	{ 10, 6,  2},
	{ 12, 6,  2},
	{ 14, 6,  2},
	{ 16, 6,  2},
	{ 5, 4,  2},
	{ 7, 4,  2},
	{ 9, 4,  2},
	{ 11, 4,  2},
	{ 13, 4,  2},
	{ 15, 4,  2},
	{ 7, 12,  1},
	{ 9, 11,  1},
	{ 11, 11,  1},
	{ 13, 12,  1},
	{ 2, 10,  1},
	{ 4, 10,  1},
	{ 6, 10,  1},
	{ 8, 9,  1},
	{ 10, 9,  1},
	{ 12, 9,  1},
	{ 14, 10,  1},
	{ 16, 10,  1},
	{ 18, 10,  1},
	{ 3, 8,  1},
	{ 3, 6,  1},
	{ 5, 8,  1},
	{ 5, 6,  1},
	{ 7, 7,  1},
	{ 9, 7,  1},
	{ 11, 7,  1},
	{ 13, 7,  1},
	{ 15, 8,  1},
	{ 17, 8,  1},
	{ 3, 4,  1},
	{ 5, 4,  1},
	{ 7, 5,  1},
	{ 9, 5,  1},
	{ 11, 5,  1},
	{ 13, 5,  1},
	{ 15, 6,  1},
	{ 17, 6,  1},
	{ 2, 2,  1},
	{ 4, 2,  1},
	{ 6, 2,  1},
	{ 8, 3,  1},
	{ 10, 3,  1},
	{ 12, 3,  1},
	{ 15, 4,  1},
	{ 17, 4,  1},
	{ 7, 0,  1},
	{ 9, 1,  1},
	{ 11, 1,  1},
	{ 14, 2,  1},
	{ 16, 2,  1},
	{ 18, 2,  1},
	{ 13, 0,  1},
	{ 6, 12,  0},
	{ 8, 12,  0},
	{ 10, 12,  0},
	{ 12, 12,  0},
	{ 14, 12,  0},
	{ 1, 11,  0},
	{ 3, 11,  0},
	{ 1, 9,  0},
	{ 0, 6,  0},
	{ 3, 9,  0},
	{ 5, 10,  0},
	{ 7, 10,  0},
	{ 9, 10,  0},
	{ 11, 10,  0},
	{ 13, 10,  0},
	{ 15, 10,  0},
	{ 17, 11,  0},
	{ 19, 11,  0},
	{ 2, 7,  0},
	{ 4, 7,  0},
	{ 6, 8,  0},
	{ 8, 8,  0},
	{ 2, 5,  0},
	{ 4, 5,  0},
	{ 6, 6,  0},
	{ 8, 6,  0},
	{ 10, 8,  0},
	{ 10, 6,  0},
	{ 12, 8,  0},
	{ 12, 6,  0},
	{ 14, 8,  0},
	{ 14, 6,  0},
	{ 17, 9,  0},
	{ 16, 7,  0},
	{ 19, 9,  0},
	{ 18, 7,  0},
	{ 1, 3,  0},
	{ 3, 3,  0},
	{ 6, 4,  0},
	{ 8, 4,  0},
	{ 10, 4,  0},
	{ 12, 4,  0},
	{ 14, 4,  0},
	{ 16, 5,  0},
	{ 18, 5,  0},
	{ 20, 6,  0},
	{ 1, 1,  0},
	{ 3, 1,  0},
	{ 5, 2,  0},
	{ 7, 2,  0},
	{ 9, 2,  0},
	{ 11, 2,  0},
	{ 13, 2,  0},
	{ 15, 2,  0},
	{ 17, 3,  0},
	{ 19, 3,  0},
	{ 17, 1,  0},
	{ 19, 1,  0},
	{ 6, 0,  0},
	{ 8, 0,  0},
	{ 10, 0,  0},
	{ 12, 0,  0},
	{ 14, 0,  0}
};

tilepos pyramid_map [MAX_TILES] = {
 {13, 14, 6}, {13, 8,  6},  {13, 2, 6}, {12, 14,  5},
 {14, 14,  5}, {12, 8,  5}, {14, 8,  5}, {12, 2,  5},
 {14, 2,  5}, {10, 14,  4}, {12, 14,  4}, {14, 14,  4},
 {16, 14,  4}, {2, 8,  4}, {11, 8,  4}, {13, 8,  4},
 {15, 8,  4}, {24, 8,  4}, {10, 2,  4}, {12, 2,  4},
 {14, 2,  4}, {16, 2,  4}, {8, 14,  3}, {10, 14,  3},
 {12, 14,  3}, {14, 14,  3}, {16, 14,  3}, {18, 14,  3},
 {2, 9,  3}, {24, 9, 3}, {10, 8,  3}, {12, 8,  3},
 {14, 8,  3}, {16, 8, 3}, {2, 7,  3}, {24, 7,  3},
 {8, 2,  3}, {10, 2,  3}, {12, 2,  3}, {14, 2,  3},
 {16, 2,  3}, {18, 2,  3}, {6, 14,  2}, {8, 14,  2},
 {10, 14,  2}, {12, 14,  2}, {14, 14,  2}, {16, 14,  2},
 {18, 14,  2}, {20, 14,  2}, {2, 10,  2}, {24, 10,  2},
 {2, 8,  2}, {9, 8,  2}, {11, 8,  2}, {13, 8,  2},
 {15, 8,  2}, {17, 8,  2}, {24, 8,  2}, {2, 6,  2},
 {24, 6,  2}, {6, 2,  2}, {8, 2,  2}, {10, 2,  2},
 {12, 2,  2}, {14, 2,  2}, {16, 2,  2}, {18, 2,  2},
 {20, 2,  2}, {4, 14,  1}, {6, 14,  1}, {8, 14,  1},
 {10, 14,  1}, {12, 14,  1}, {14, 14,  1}, {16, 14,  1},
 {18, 14,  1}, {20, 14,  1}, {22, 14,  1}, {2, 11,  1},
 {24, 11,  1}, {2, 9,  1}, {24, 9,  1}, {8, 8,  1},
 {10, 8,  1}, {12, 8,  1}, {14, 8,  1}, {16, 8,  1},
 {18, 8,  1}, {2, 7,  1}, {24, 7,  1}, {2, 5,  1},
 {24, 5,  1}, {4, 2,  1}, {6, 2,  1}, {8, 2,  1},
 {10, 2,  1}, {12, 2,  1}, {14, 2,  1}, {16, 2,  1},
 {18, 2,  1}, {20, 2,  1}, {22, 2,  1}, {2, 14,  0},
 {4, 14,  0}, {6, 14,  0}, {8, 14,  0}, {10, 14,  0},
 {12, 14,  0}, {14, 14,  0}, {16, 14,  0}, {18, 14,  0},
 {20, 14,  0}, {22, 14,  0}, {24, 14,  0}, {2, 12,  0},
 {24, 12,  0}, {2, 10,  0}, {24, 10,  0}, {2, 8, 0},
 {7, 8,  0}, {9, 8,  0}, {11, 8,  0}, {13, 8,  0},
 {15, 8,  0}, {17, 8,  0}, {19, 8,  0}, {24, 8, 0},
 {2, 6,  0}, {24, 6,  0}, {2, 4,  0}, {24, 4,  0},
 {2, 2,  0}, {4, 2,  0}, {6, 2,  0}, {8, 2,  0},
 {10, 2,  0}, {12, 2,  0}, {14, 2,  0}, {16, 2,  0},
 {18, 2,  0}, {20, 2,  0}, {22, 2,  0}, {24, 2,  0}
};

tilepos cross_map [MAX_TILES] = {
 {13, 8, 5},
 {13, 12,  4}, {13, 10,  4}, {9, 8,  4}, {11, 8,  4},
 {13, 8,  4}, {15, 8,  4}, {17, 8,  4}, {13, 6,  4},
 {13, 4,  4}, {5, 14,  3}, {21, 14,  3}, {13, 13,  3},
 {13, 11,  3}, {6, 8,  3},
 {8, 8, 3}, {10, 8,  3}, {12, 9,  3}, {12, 7,  3}, {14, 9,  3}, {14, 7,  3},  {16, 8,  3}, {18, 8,  3}, {20, 8, 3},
   {13, 5,  3},
 {13, 3,  3}, {5, 2,  3}, {21, 2,  3}, {5, 14,  2},
 {7, 14,  2}, {13, 14,  2}, {19, 14,  2}, {21, 14,  2},
 {5, 12,  2}, {13, 12,  2}, {21, 12,  2}, {13, 10, 2},
 {5, 8, 2}, {7, 8,  2}, {9, 8,  2}, {11, 8,  2},
 {13, 8,  2}, {15, 8,  2}, {17, 8,  2}, {19, 8,  2},
 {21, 8, 2}, {13, 6, 2}, {5, 4,  2}, {13, 4,  2},
 {21, 4,  2}, {5, 2,  2}, {7, 2,  2}, {13, 2,  2},
 {19, 2,  2}, {21, 2,  2}, {13, 15,  1}, {5, 14,  1},
 {7, 14,  1}, {9, 14,  1}, {17, 14,  1}, {19, 14,  1},
 {21, 14,  1}, {13, 13,  1}, {5, 12,  1}, {7, 12, 1},
 {19, 12, 1}, {21, 12,  1}, {13, 11,  1}, {5, 10,  1},
 {21, 10,  1}, {4, 8,  1},
 {6, 8,  1}, {8, 8,  1}, {10, 8,  1}, {12, 9,  1}, {12, 7,  1},
 {14, 9,  1}, {14, 7,  1},  {16, 8,  1}, {18, 8,  1}, {20, 8,  1},
 {22, 8,  1}, {5, 6,  1}, {21, 6,  1}, {13, 5,  1},
 {5, 4,  1}, {7, 4, 1}, {19, 4, 1}, {21, 4,  1},
 {13, 3,  1}, {5, 2,  1}, {7, 2,  1}, {9, 2,  1},
 {17, 2,  1},{19, 2,  1}, {21, 2,  1}, {13, 1,  1},
 {13, 16,  0}, {5, 14,  0}, {7, 14,  0}, {9, 14,  0},
 {11, 14,  0}, {13, 14,  0}, {15, 14,  0}, {17, 14,  0},
 {19, 14,  0}, {21, 14,  0}, {5, 12,  0}, {7, 12, 0},
 {13, 12,  0}, {19, 12, 0}, {21, 12,  0}, {5, 10,  0},
 {13, 10,  0}, {21, 10,  0}, {3, 8, 0},{5, 8, 0},
 {7, 8,  0}, {9, 8,  0}, {11, 8,  0}, {13, 8,  0},
 {15, 8,  0}, {17, 8,  0}, {19, 8,  0}, {21, 8, 0},
 {23, 8,  0}, {5, 6,  0}, {13, 6,  0}, {21, 6,  0},
 {5, 4,  0}, {7, 4, 0}, {13, 4,  0}, {19, 4, 0},
 {21, 4,  0}, {5, 2,  0}, {7, 2,  0}, {9, 2,  0},
 {11, 2,  0}, {13, 2,  0}, {15, 2,  0}, {17, 2,  0},
 {19, 2,  0}, {21, 2,  0}, {13, 0, 0}
};


tilepos tictactoe_map [MAX_TILES] = {
 {6, 12,  4},{8, 12,  4}, {10, 12,  4}, {12, 12,  4}, {14, 12,  4}, {16, 12, 4},{18, 12,  4},
 {6, 10,  4}, {18, 10,  4},
 {6, 8,  4}, {18, 8,  4},
 {6, 6,  4}, {18, 6,  4},
 {6, 4,  4}, {8, 4,  4}, {10, 4,  4}, {12, 4,  4}, {14, 4,  4}, {16, 4, 4},{18, 4,  4},

 {6, 12,  3}, {8, 12,  3}, {10, 12,  3}, {12, 12,  3}, {14, 12,  3}, {16, 12, 3}, {18, 12,  3},
 {6, 10,  3}, {18, 10,  3},
 {6, 8,  3}, {18, 8,  3},
 {6, 6,  3}, {18, 6,  3},
 {6, 4,  3}, {8, 4,  3}, {10, 4,  3}, {12, 4,  3}, {14, 4,  3}, {16, 4,  3}, {18, 4,  3},

 {6, 14,  2}, {18, 14,  2},
 {4, 12,  2}, {6, 12,  2}, {8, 12,  2}, {10, 12,  2}, {12, 12,  2}, {14, 12, 2}, {16, 12,  2}, {18, 12,  2}, {20, 12,  2},
 {6, 10,  2}, {18, 10,  2},
 {6, 8,  2}, {18, 8,  2},
 {6, 6,  2}, {18, 6,  2},
 {4, 4,  2}, {6, 4,  2}, {8, 4,  2}, {10, 4,  2}, {12, 4,  2}, {14, 4,  2}, {16, 4,  2}, {18, 4,  2}, {20, 4,  2},
 {6, 2,  2}, {18, 2,  2},

 {6, 16,  1}, {18, 16,  1},
 {6, 14,  1}, {18, 14,  1},
 {2, 12,  1}, {4, 12,  1}, {6, 12,  1}, {8, 12,  1}, {10, 12,  1}, {12, 12, 1}, {14, 12,  1}, {16, 12,  1}, {18, 12,  1}, {20, 12,  1}, {22, 12,  1},
 {6, 10,  1}, {18, 10,  1},
 {6, 8,  1}, {18, 8,  1},
 {6, 6,  1}, {18, 6,  1},
 {2, 4,  1}, {4, 4,  1}, {6, 4,  1}, {8, 4,  1}, {10, 4,  1}, {12, 4,  1}, {14, 4,  1}, {16, 4,  1}, {18, 4,  1}, {20, 4,  1}, {22, 4,  1},
 {6, 2,  1}, {18, 2,  1},
 {6, 0,  1}, {18, 0,  1},

 {6, 16,  0}, {18, 16,  0},
 {6, 14,  0}, {18, 14,  0},
 {0, 12, 0}, {2, 12,  0}, {4, 12,  0}, {6, 12,  0}, {8, 12,  0}, {10, 12,  0}, {12, 12,  0}, {14, 12,  0}, {16, 12,  0},{18, 12,  0}, {20, 12,  0},{22, 12, 0}, {24, 12, 0},
 {6, 10,  0}, {18, 10,  0},
 {6, 8,  0}, {18, 8,  0},
 {6, 6,  0}, {18, 6,  0},
 {0, 4, 0}, {2, 4,  0}, {4, 4,  0}, {6, 4,  0}, {8, 4,  0}, {10, 4,  0}, {12, 4,  0}, {14, 4,  0}, {16, 4,  0},{18, 4,  0}, {20, 4,  0},{22, 4,  0}, {24, 4, 0},
 {6, 2,  0}, {18, 2,  0},
 {6, 0,  0}, {18, 0,  0},
};

tilepos cloud_map [MAX_TILES] = {
 {12, 11,  2},

 {0, 6, 2}, {4, 6, 2}, {8, 6, 2}, {12, 6, 2}, {16, 6, 2}, {20, 6, 2}, {24, 6, 2},
 {0, 4, 2}, {4, 4, 2}, {8, 4, 2}, {12, 4, 2}, {16, 4, 2}, {20, 4, 2}, {24, 4, 2},
 {0, 2, 2}, {4, 2, 2}, {8, 2, 2}, {12, 2, 2}, {16, 2, 2}, {20, 2, 2}, {24, 2, 2},
 {0, 0, 2}, {4, 0, 2}, {8, 0, 2}, {12, 0, 2}, {16, 0, 2}, {20, 0, 2}, {24, 0, 2},

 {6, 11, 1}, {8, 11, 1}, {10, 11, 1}, {12, 11, 1}, {14, 11, 1},
 {16, 11, 1}, {18, 11, 1}, {20, 11, 1},

 {0, 6, 1}, {4, 6, 1}, {8, 6, 1}, {12, 6, 1}, {16, 6, 1}, {20, 6, 1}, {24, 6, 1},
 {0, 4, 1}, {4, 4, 1}, {8, 4, 1}, {12, 4, 1}, {16, 4, 1}, {20, 4, 1}, {24, 4, 1},
 {0, 2, 1}, {4, 2, 1}, {8, 2, 1}, {12, 2, 1}, {16, 2, 1}, {20, 2, 1}, {24, 2, 1},
 {0, 0, 1}, {4, 0, 1}, {8, 0, 1}, {12, 0, 1}, {16, 0, 1}, {20, 0, 1}, {24, 0, 1},

 {5, 11, 0}, {7, 11, 0}, {9, 11, 0}, {11, 11, 0}, {13, 11, 0},
 {15, 11, 0}, {17, 11, 0}, {19, 11, 0}, {21, 11, 0},

 {0, 8, 0}, {2, 8, 0}, {4, 8, 0}, {6, 8, 0}, {8, 8, 0}, {10, 8, 0}, {12, 8, 0},
 {14, 8, 0}, {16, 8, 0}, {18, 8, 0}, {20, 8, 0}, {22, 8, 0}, {24, 8, 0}, {26, 8, 0},

 {0, 6, 0}, {2, 6, 0}, {4, 6, 0}, {6, 6, 0}, {8, 6, 0}, {10, 6, 0}, {12, 6, 0},
 {14, 6, 0}, {16, 6, 0}, {18, 6, 0}, {20, 6, 0}, {22, 6, 0}, {24, 6, 0}, {26, 6, 0},

 {0, 4, 0}, {2, 4, 0}, {4, 4, 0}, {6, 4, 0}, {8, 4, 0}, {10, 4, 0}, {12, 4, 0},
 {14, 4, 0}, {16, 4, 0}, {18, 4, 0}, {20, 4, 0}, {22, 4, 0}, {24, 4, 0}, {26, 4, 0},

 {0, 2, 0}, {2, 2, 0}, {4, 2, 0}, {6, 2, 0}, {8, 2, 0}, {10, 2, 0}, {12, 2, 0},
 {14, 2, 0}, {16, 2, 0}, {18, 2, 0}, {20, 2, 0}, {22, 2, 0}, {24, 2, 0}, {26, 2, 0},

 {0, 0, 0}, {2, 0, 0}, {4, 0, 0}, {6, 0, 0}, {8, 0, 0}, {10, 0, 0}, {12, 0, 0},
 {14, 0, 0}, {16, 0, 0}, {18, 0, 0}, {20, 0, 0}, {22, 0, 0}, {24, 0, 0}, {26, 0, 0},

};

tilepos reddragon_map [MAX_TILES] = {
{10,8,2},
{12,8,2},
{14,8,2},
{16,8,2},
{22,8,2},
{10,6,2},
{12,6,2},
{14,6,2},
{16,6,2},
{10,4,2},
{12,4,2},
{14,4,2},
{16,4,2},
{10,2,2},
{12,2,2},
{14,2,2},
{16,2,2},
{5,9,1},
{7,9,1},
{9,9,1},
{11,9,1},
{13,9,1},
{15,9,1},
{17,9,1},
{19,9,1},
{21,9,1},
{7,7,1},
{9,7,1},
{11,7,1},
{13,7,1},
{15,7,1},
{17,7,1},
{19,7,1},
{21,7,1},
{23,7,1},
{7,5,1},
{9,5,1},
{11,5,1},
{13,5,1},
{15,5,1},
{17,5,1},
{19,5,1},
{21,5,1},
{23,5,1},
{5,3,1},
{7,3,1},
{9,3,1},
{11,3,1},
{13,3,1},
{15,3,1},
{17,3,1},
{19,3,1},
{21,3,1},
{23,3,1},
{7,1,1},
{9,1,1},
{11,1,1},
{13,1,1},
{15,1,1},
{17,1,1},
{19,1,1},
{21,1,1},
{4,13,0},
{8,13,0},
{12,13,0},
{16,13,0},
{20,13,0},
{24,13,0},
{0,12,0},
{28,12,0},
{4,10,0},
{6,10,0},
{8,10,0},
{10,10,0},
{12,10,0},
{14,10,0},
{16,10,0},
{18,10,0},
{20,10,0},
{22,10,0},
{24,10,0},
{0,9,0},
{28,9,0},
{4,8,0},
{6,8,0},
{8,8,0},
{10,8,0},
{12,8,0},
{14,8,0},
{16,8,0},
{18,8,0},
{20,8,0},
{22,8,0},
{24,8,0},
{0,6,0},
{4,6,0},
{6,6,0},
{8,6,0},
{10,6,0},
{12,6,0},
{14,6,0},
{16,6,0},
{18,6,0},
{20,6,0},
{22,6,0},
{24,6,0},
{28,6,0},
{4,4,0},
{6,4,0},
{8,4,0},
{10,4,0},
{12,4,0},
{14,4,0},
{16,4,0},
{18,4,0},
{20,4,0},
{22,4,0},
{24,4,0},
{0,3,0},
{28,3,0},
{4,2,0},
{6,2,0},
{8,2,0},
{10,2,0},
{12,2,0},
{14,2,0},
{16,2,0},
{18,2,0},
{20,2,0},
{22,2,0},
{24,2,0},
{0,0,0},
{4,0,0},
{6,0,0},
{8,0,0},
{10,0,0},
{12,0,0},
{14,0,0},
{16,0,0},
{18,0,0},
{20,0,0},
{22,0,0},
{24,0,0},
{28,0,0}
};

/* FIXME: The games loaded from file won't have translated names. */

map hardcoded[] = {
	{ N_("Easy"), "easy", easy_map, TRUE, 0, 0 },
	{ N_("Difficult"), "difficult", hard_map, TRUE, 0, 0 },
        { N_("Confounding Cross"), "confounding", cross_map, TRUE, 0, 0 },
        { N_("Pyramid's Walls"), "pyramid", pyramid_map, TRUE, 0, 0 },
        { N_("Tic-Tac-Toe"), "tictactoe", tictactoe_map, TRUE, 0, 0 },
        { N_("Cloud"), "cloud", cloud_map, TRUE, 0, 0 },
        { N_("Red Dragon"), "dragon", reddragon_map, TRUE, 0, 0 },
#if 0
        { N_("Four Bridges"), "bridges", fourbridges_map, TRUE, 0, 0 },
	{ N_("The Ziggurat"), "ziggurat", ziggurat_map, TRUE, 0, 0 }, 
#endif
};


GList *maplist;

#define TAG_NAME_MAHJONGG "mahjongg"
#define TAG_NAME_MAP "map"
#define TAG_NAME_LAYER "layer"
#define TAG_NAME_ROW "row"
#define TAG_NAME_COLUMN "column"
#define TAG_NAME_BLOCK "block"
#define TAG_NAME_TILE "tile"

#define ATTR_NAME "name"
#define ATTR_SCORENAME "scorename"
#define ATTR_LEFT "left"
#define ATTR_RIGHT "right"
#define ATTR_TOP "top"
#define ATTR_BOTTOM "bottom"
#define ATTR_X "x"
#define ATTR_Y "y"
#define ATTR_Z "z"

tilepos testmap[MAX_TILES];

gboolean in_mahjongg;

gint x1, x2, y1, y2, z;
const gchar *name, *score_name;

enum {
  ATTR_INT,
  ATTR_HALFINT,
  ATTR_STRING,
  ATTR_END
};

struct attrs {
  gchar *name;
  gint length;
  gint type;
  void *value;
} attributes[] = { {ATTR_NAME, sizeof (ATTR_NAME), ATTR_STRING, &name},
		   {ATTR_SCORENAME, sizeof (ATTR_SCORENAME), 
		    ATTR_STRING, &score_name},
		   {ATTR_LEFT, sizeof (ATTR_LEFT), ATTR_HALFINT, &x1},
		   {ATTR_RIGHT, sizeof (ATTR_RIGHT), ATTR_HALFINT, &x2},
		   {ATTR_TOP, sizeof (ATTR_TOP), ATTR_HALFINT, &y1},
		   {ATTR_BOTTOM, sizeof (ATTR_BOTTOM), ATTR_HALFINT, &y2},
		   {ATTR_X, sizeof (ATTR_X), ATTR_HALFINT, &x1},
		   {ATTR_Y, sizeof (ATTR_Y), ATTR_HALFINT, &y1},
		   {ATTR_Z, sizeof (ATTR_Z), ATTR_INT, &z},
		   {NULL, 0, ATTR_END, NULL}};
struct attrs *a;

enum {
  TAG_MAHJONGG,
  TAG_MAP,
  TAG_LAYER,
  TAG_BLOCK,
  TAG_ROW,
  TAG_COLUMN,
  TAG_TILE,
  TAG_END
};

struct tgs {
  gchar *name;
  gint length;
  gint value;
} tags[] = { {TAG_NAME_MAHJONGG, sizeof (TAG_NAME_MAHJONGG), TAG_MAHJONGG},
	     {TAG_NAME_MAP, sizeof (TAG_NAME_MAP), TAG_MAP},
	     {TAG_NAME_LAYER, sizeof (TAG_NAME_LAYER), TAG_LAYER},
	     {TAG_NAME_BLOCK, sizeof (TAG_NAME_BLOCK), TAG_BLOCK},
	     {TAG_NAME_ROW, sizeof (TAG_NAME_ROW), TAG_ROW},
	     {TAG_NAME_COLUMN, sizeof (TAG_NAME_COLUMN), TAG_COLUMN},
	     {TAG_NAME_TILE, sizeof (TAG_NAME_TILE), TAG_TILE},
	     {NULL, TAG_END}};
struct tgs *t;	       

GMarkupParser parser;

map *maps = NULL;
gint nmaps = 0;

static void parse_start_el (GMarkupParseContext *context,
			    const gchar         *element_name,
			    const gchar        **attribute_names,
			    const gchar        **attribute_values,
			    gpointer             user_data,
			    GError             **error)
{
  map *m;
  gint tag;
  gint y;

  /* Identify the tag. */
  tag = TAG_END;
  t = tags;
  while (t->value != TAG_END) {
    if (g_ascii_strncasecmp (element_name, t->name, t->length) == 0) {
      tag = t->value;
      break;
    }
    t++;
  }

  /* Ignore anything outside of a pair of <mahjongg> tags. */
  if (!in_mahjongg) {
    in_mahjongg = tag == TAG_MAHJONGG;
    return;
  }

  /* Scan the attributes. */
  x1 = x2 = y1 = y2 = z = 0;
  name = score_name = NULL;

  m = NULL;

  if (maplist) {
    m = maplist->data;
    z = m->layer_t;
  }

  while (*attribute_names) {
    a = attributes;
    while (a->type != ATTR_END) {
      if (g_ascii_strncasecmp (*attribute_names, a->name, a->length) == 0) {
	switch (a->type) {
	case ATTR_INT:
	  *(gint *)(a->value) = (gint)(atof (*attribute_values));
	  break;	  
	case ATTR_HALFINT:
	  *(gint *)(a->value) = (gint)(2*atof (*attribute_values));
	  break;
	case ATTR_STRING:
	  *(const gchar **)(a->value) = *attribute_values;
	  break;
	}
	break;
      } 
      a++;
    }

    attribute_names++;
    attribute_values++;
  }

  /* Handle setting up the maplist separately. */
  if (tag == TAG_MAP) {
    m = g_malloc0 (sizeof (map));
    maplist = g_list_prepend (maplist, m);
    m->map = g_malloc0 (sizeof (tilepos) * MAX_TILES);
    m->current = 0;
    m->hardcoded = FALSE;
    m->name = g_strdup (name);
    m->score_name = g_strdup (score_name);
    return;
  }

  /* This only happens if the input file is malformed. */
  if (!maplist)
    return;
  
  /* Now handle the rest of the tags. */
  switch (tag) {
  case TAG_LAYER:
    m->layer_t = CLAMP (z, 0, 255);
    break;
  case TAG_ROW:
    for ( ; x1 <= x2; x1 += 2) {
      m->map[m->current].x = x1;
      m->map[m->current].y = y1;
      m->map[m->current].layer = z;
      if (m->current < MAX_TILES) 
	m->current++;
    }
    break;
  case TAG_COLUMN:
    for ( ; y1 <= y2; y1 += 2) {
      m->map[m->current].x = x1;
      m->map[m->current].y = y1;
      m->map[m->current].layer = z;
      if (m->current < MAX_TILES) 
	m->current++;
    }
    break;
  case TAG_BLOCK:
    for ( ; x1 <= x2; x1 += 2) {
      for (y = y1; y <= y2; y += 2) {
	m->map[m->current].x = x1;
	m->map[m->current].y = y;
	m->map[m->current].layer = z;
	if (m->current < MAX_TILES) 
	  m->current++;
      }
    }
    break;
  case TAG_TILE:
    m->map[m->current].x = x1;
    m->map[m->current].y = y1;
    m->map[m->current].layer = z;
    if (m->current < MAX_TILES) 
      m->current++;
    break;
  }

}

/* Sorted such that the bottom leftest are first, and layers decrease
 * Bottom left = high y, low x ! */
static int compare_tiles (tilepos *a, tilepos *b)
{
  int dx, dy;

  dx = b->x - a->x;
  dy = b->y - a->y;

  if (b->layer < a->layer)
    return -1;
  if (b->layer > a->layer)
    return +1;
  if (dx > dy)
    return -1;
  if (dx < dy)
    return +1;

  return 0;
}

static void parse_end_el (GMarkupParseContext *context,
			    const gchar         *element_name,
			    gpointer             user_data,
			    GError             **error)
{
  map *m;

  if (!in_mahjongg)
    return;

  /* FIXME: What if we end mahjongg without closing <map> ? */

  if (g_ascii_strncasecmp (TAG_NAME_MAHJONGG, element_name, 
			   sizeof (TAG_NAME_MAHJONGG)) == 0) {
    in_mahjongg = FALSE;
    return;
  }

  if (g_ascii_strncasecmp (TAG_NAME_MAP, element_name, 
			   sizeof (TAG_NAME_MAP)) == 0) {
    m = maplist->data;
    if (m->current < MAX_TILES)
      g_warning ("Too few tiles in map %s", m->name);
    qsort (m->map, MAX_TILES, sizeof (tilepos), 
	   (int (*)(const void *, const void *)) compare_tiles);
  }
}

static void load_map_from_file (gchar *filename)
{
  GMarkupParseContext *parse_context;
  gboolean ok;
  gchar *file;
  gint length;

  /* Read in a map file. */
  ok = g_file_get_contents (filename, &file, &length, NULL);

  if (!ok) {
    g_warning ("Could not read file %s\n", filename);
    return;
  }

  /* Parse the map file. */
  in_mahjongg = FALSE;

  parse_context = g_markup_parse_context_new (&parser, 0, NULL, NULL);
  g_markup_parse_context_parse (parse_context, file, length, NULL);

  /* FIXME: Handle bad maps (i.e. throw them off the list) */
  /* FIXME: Check for duplicates. */
}

void load_maps (void)
{
  int i;
  map *m;
  GamesFileList *filelist;

  /* Initialise the parser. */
  parser.start_element = parse_start_el;
  parser.end_element = parse_end_el;
  parser.text = NULL;
  parser.passthrough = NULL;
  parser.error = NULL;

  maplist = NULL;

  /* Load up the hard-coded games. */
  for (i=0; i<G_N_ELEMENTS (hardcoded); i++) {
    maplist = g_list_prepend (maplist, hardcoded + i);
  }

  /* Now read the remainder in from file. */
  filelist = games_file_list_new ("*.map", ".", MAPDIR, NULL);
  games_file_list_for_each (filelist, (GFunc)load_map_from_file, NULL);

  /* FIXME: Ideally we would do this transformation, but the old code 
   * expects an array, so we give it one. */
  if (maps)
    g_free (maps);
  
  nmaps = g_list_length (maplist);

  maps = g_malloc0 (sizeof (map)*nmaps);

  for (i = 0; i<nmaps; i++) {
    m = maplist->data;
    g_memmove (maps + i, maplist->data, sizeof (map));
    if (!m->hardcoded)
      g_free (maplist->data);
    maplist = g_list_delete_link (maplist, maplist);
  }
  
}

